package tourplaner.business;

import org.apache.log4j.Logger;
import tourplaner.data.DbConnect;
import tourplaner.object.Log;
import tourplaner.object.Tour;

import java.awt.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Haupt Logik des Tourplaners
 */
public class TourPlaner{
    private Logger logger;

    /**
     * Erstellt alle DB Tabellen falls nötig
     */
    public static void init(){
        new DbConnect().init();
    }

    /**
     * Holt alle touren aus der DB 
     * @return Alle touren als ArrayList
     */
    public static ArrayList<Tour> getAllTours(){
        return new DbConnect().getAllTouren();
    }

    /**
     * Bearbeitet eine Tour
     * @param oldname Alter Tourname
     * @param tour Neuer Tourname
     * @return false bei error
     */
    public static boolean editTour(String oldname, Tour tour) throws IOException {
        FileHelper.delFile(new File(getImagePath(oldname)));
        FileHelper.delFile(new File(getImagePdfPath(oldname)));
        DirectionMap directionMap = new DirectionMap(tour.getStart(), tour.getZiel(), tour.getName());
        tour.setDauer(directionMap.getDauer()+"");
        tour.setStrecke(directionMap.getStrecke());
        return new DbConnect().editTour(oldname, tour);
    }

    /**
     * Löscht eine Tour
     * @param tourname Tourname
     * @return false bei error
     */
    public static boolean delTour(String tourname){
        FileHelper.delFile(new File(getImagePath(tourname)));
        FileHelper.delFile(new File(getImagePdfPath(tourname)));
        return new DbConnect().delTour(tourname);
    }
    /**
     * Fügt eine Tour hinzu
     * @param newTour Neue Tour
     * @return false bei error
     */
    public static boolean addTour(Tour newTour) throws IOException {
        DirectionMap directionMap = new DirectionMap(newTour.getStart(), newTour.getZiel(), newTour.getName());
        newTour.setDauer(directionMap.getDauer()+"");
        newTour.setStrecke(directionMap.getStrecke());
        return new DbConnect().addTour(newTour);
    }

    /**
     * Holt das Gui Bild
     * @param tourname Name der Tour
     * @return Das Bild
     * @throws IOException Error beim Bild holen
     */
    public static Image getImage(String tourname) throws IOException {
        return FileHelper.getImage(new File(getImagePath(tourname)));
    }

     /**
     * Path zu dem Bild für die GUI
     * @param tourname Name der Tour
     * @return Path zu dem Bild für die GUI
     */
    public static String getImagePath(String tourname){
        return ConfigHelper.getIniString(ConfigHelper.getStandartConfig(), "map", "path") + tourname + ".jpg";
    }

    /**
     * Path zu dem Bild für die PDF
     * @param tourname Name der Tour
     * @return Path zu dem Bild für die PDF
     */
    public static String getImagePdfPath(String tourname){
        return ConfigHelper.getIniString(ConfigHelper.getStandartConfig(), "map", "path") + tourname + "_pdf.jpg";
    }

    /**
     * Öffnet ein Bild einer Tour
     * @param tourname Tourname
     */
    public static void openImage(String tourname){
        FileHelper.openDefault(getImagePath(tourname));
    }


    /**
     * Holt alle Logs einer Tour
     * @param tourname Tourname
     * @return Alle Logs der Tour
     */
    public static ArrayList<Log> getLogs(String tourname){
        return new DbConnect().getLogs(tourname);
    }

    /**
     * Führt einen Rport aus
     * @param tourname Name der Tour die gereportet wird
     */
    public static void doReport(String tourname){
        Reporter.createTourReport(tourname);
    }

    /**
     * Holt eine Tour
     * @param tourname Name der zu holenden Tour
     * @return Tour die geholt werden soll
     */
    public static Tour getTour(String tourname){
        return new DbConnect().getTour(tourname);
    }

    /**
     * Fügt ein Log hinzu
     * @param tourname Name der Tour
     * @param log Neues Log
     * @return false bei error
     */
    public static boolean addLog(String tourname, Log log){
        if(Double.isInfinite(log.getAvgspeed())) log.setAvgspeed(-1.0);
        return new DbConnect().addLog(tourname, log);
    }

    /**
     * Löscht ein Log
     * @param tourname Name der Tour
     * @param id Id des Logs
     * @return false bei error
     */
    public static boolean delLog(String tourname, String id){
        return new DbConnect().delLog(tourname, id);
    }

    /**
     * Editiert ein Log
     * @param tourname Name der Tour
     * @param log Neues Log
     * @return false bei error
     */
    public static boolean editLog(String tourname, Log log){
        if(Double.isInfinite(log.getAvgspeed())) log.setAvgspeed(-1.0);
        return new DbConnect().editLog(tourname, log);
    }

    /**
     * Exportiert die Daten
     * @param path Path zu den Daten
     * @throws IOException Fehler beim schreiben der Daten
     */
    public static void exportData(String path) throws IOException {
            Exporter exporter = new Exporter(path);
            exporter.doExport();
            FileHelper.openDefault(path);
    }

    /**
     * Importiert daten
     * @param path Path zu den Daten
     * @throws FileNotFoundException File konnte nicht geöffnet werden
     */
    public static void importData(String path) throws FileNotFoundException {
            Exporter exporter = new Exporter(path);
            exporter.doImport();
    }
}
